@@ -38,7 +38,7 @@
 /*
   Compare ASCII string against the string with the specified
   character set.
-  Only compares the equality, case insencitive.
+  Only compares the equality, case insensitive.
 */
 static bool eq_ascii_string(const CHARSET_INFO *cs,
                             const char *ascii,
@@ -133,7 +133,7 @@ static const char tab_arr[TAB_SIZE_LIMIT+1]= "        ";
 
 static int append_tab(String *js, int depth, int tab_size)
 {
-  if (js->append("\n", 1))
+  if (js->append('\n'))
     return 1;
   for (int i=0; i<depth; i++)
   {
@@ -257,16 +257,21 @@ static int json_nice(json_engine_t *je, String *nice_js,
                      Item_func_json_format::formats mode, int tab_size=4)
 {
   int depth= 0;
-  const char *comma, *colon;
+  static const char *comma= ", ", *colon= "\": ";
   uint comma_len, colon_len;
   int first_value= 1;
+  int value_size = 0;
+  int curr_state= -1;
+  int64_t value_len= 0;
+  String curr_str{};
+
+  nice_js->length(0);
+  nice_js->set_charset(je->s.cs);
+  nice_js->alloc(je->s.str_end - je->s.c_str + 32);
 
-  DBUG_ASSERT(je->s.cs == nice_js->charset());
   DBUG_ASSERT(mode != Item_func_json_format::DETAILED ||
               (tab_size >= 0 && tab_size <= TAB_SIZE_LIMIT));
 
-  comma= ", ";
-  colon= "\": ";
   if (mode == Item_func_json_format::LOOSE)
   {
     comma_len= 2;
@@ -285,6 +290,7 @@ static int json_nice(json_engine_t *je, String *nice_js,
 
   do
   {
+    curr_state= je->state;
     switch (je->state)
     {
     case JST_KEY:
@@ -307,7 +313,7 @@ static int json_nice(json_engine_t *je, String *nice_js,
             append_tab(nice_js, depth, tab_size))
           goto error;
 
-        nice_js->append("\"", 1);
+        nice_js->append('"');
         append_simple(nice_js, key_start, key_end - key_start);
         nice_js->append(colon, colon_len);
       }
@@ -332,17 +338,23 @@ handle_value:
         if (append_simple(nice_js, je->value_begin,
                           je->value_end - je->value_begin))
           goto error;
-
+        
+        curr_str.copy((const char *)je->value_begin,
+                      je->value_end - je->value_begin, je->s.cs);
+        value_len= je->value_end - je->value_begin;
         first_value= 0;
+        if (value_size != -1)
+          value_size++;
       }
       else
       {
         if (mode == Item_func_json_format::DETAILED &&
-            depth > 0 &&
+            depth > 0 && !(curr_state != JST_KEY) &&
             append_tab(nice_js, depth, tab_size))
           goto error;
         nice_js->append((je->value_type == JSON_VALUE_OBJECT) ? "{" : "[", 1);
         first_value= 1;
+        value_size= (je->value_type == JSON_VALUE_OBJECT) ? -1: 0;
         depth++;
       }
 
@@ -351,11 +363,27 @@ handle_value:
     case JST_OBJ_END:
     case JST_ARRAY_END:
       depth--;
-      if (mode == Item_func_json_format::DETAILED &&
+      if (mode == Item_func_json_format::DETAILED && (value_size > 1 || value_size == -1) &&
           append_tab(nice_js, depth, tab_size))
         goto error;
+        
+      if (mode == Item_func_json_format::DETAILED && 
+          value_size == 1 && je->state != JST_OBJ_END)
+      {
+        for (auto i = 0; i < value_len; i++)
+        {
+          nice_js->chop();
+        }
+        for (auto i = 0; i < (depth + 1) * tab_size + 1; i++)
+        {
+          nice_js->chop();
+        }
+        nice_js->append(curr_str);
+      }
+      
       nice_js->append((je->state == JST_OBJ_END) ? "}": "]", 1);
       first_value= 0;
+      value_size= -1;
       break;
 
     default:
@@ -363,7 +391,7 @@ handle_value:
     };
   } while (json_scan_next(je) == 0);
 
-  return je->s.error;
+  return je->s.error || *je->killed_ptr;
 
 error:
   return 1;
@@ -371,15 +399,15 @@ error:
 
 
 #define report_json_error(js, je, n_param) \
-  report_json_error_ex(js, je, func_name(), n_param, \
+  report_json_error_ex(js->ptr(), je, func_name(), n_param, \
       Sql_condition::WARN_LEVEL_WARN)
 
-void report_json_error_ex(String *js, json_engine_t *je,
+void report_json_error_ex(const char *js, json_engine_t *je,
                           const char *fname, int n_param,
                           Sql_condition::enum_warning_level lv)
 {
   THD *thd= current_thd;
-  int position= (int)((const char *) je->s.c_str - js->ptr());
+  int position= (int)((const char *) je->s.c_str - js);
   uint code;
 
   n_param++;
@@ -409,16 +437,22 @@ void report_json_error_ex(String *js, json_engine_t *je,
 
   case JE_DEPTH:
     code= ER_JSON_DEPTH;
-    push_warning_printf(thd, lv, code, ER_THD(thd, code), JSON_DEPTH_LIMIT,
-                        n_param, fname, position);
+    if (lv == Sql_condition::WARN_LEVEL_ERROR)
+      my_error(code, MYF(0), JSON_DEPTH_LIMIT, n_param, fname, position);
+    else
+      push_warning_printf(thd, lv, code, ER_THD(thd, code), JSON_DEPTH_LIMIT,
+                          n_param, fname, position);
     return;
 
   default:
     return;
   }
 
-  push_warning_printf(thd, lv, code, ER_THD(thd, code),
-                      n_param, fname, position);
+  if (lv == Sql_condition::WARN_LEVEL_ERROR)
+    my_error(code, MYF(0), n_param, fname, position);
+  else
+    push_warning_printf(thd, lv, code, ER_THD(thd, code),
+                        n_param, fname, position);
 }
 
 
@@ -428,15 +462,15 @@ void report_json_error_ex(String *js, json_engine_t *je,
 #define TRIVIAL_PATH_NOT_ALLOWED 3
 
 #define report_path_error(js, je, n_param) \
-  report_path_error_ex(js, je, func_name(), n_param,\
+  report_path_error_ex(js->ptr(), je, func_name(), n_param,\
       Sql_condition::WARN_LEVEL_WARN)
 
-static void report_path_error_ex(String *ps, json_path_t *p,
-                                 const char *fname, int n_param,
-                                 Sql_condition::enum_warning_level lv)
+void report_path_error_ex(const char *ps, json_path_t *p,
+                          const char *fname, int n_param,
+                          Sql_condition::enum_warning_level lv)
 {
   THD *thd= current_thd;
-  int position= (int)((const char *) p->s.c_str - ps->ptr() + 1);
+  int position= (int)((const char *) p->s.c_str - ps + 1);
   uint code;
 
   n_param++;
@@ -455,8 +489,11 @@ static void report_path_error_ex(String *ps, json_path_t *p,
 
   case JE_DEPTH:
     code= ER_JSON_PATH_DEPTH;
-    push_warning_printf(thd, lv, code, ER_THD(thd, code),
-                        JSON_DEPTH_LIMIT, n_param, fname, position);
+    if (lv == Sql_condition::WARN_LEVEL_ERROR)
+      my_error(code, MYF(0), JSON_DEPTH_LIMIT, n_param, fname, position);
+    else
+      push_warning_printf(thd, lv, code, ER_THD(thd, code),
+                          JSON_DEPTH_LIMIT, n_param, fname, position);
     return;
 
   case NO_WILDCARD_ALLOWED:
@@ -471,12 +508,14 @@ static void report_path_error_ex(String *ps, json_path_t *p,
   default:
     return;
   }
-  push_warning_printf(thd, lv, code, ER_THD(thd, code),
-                      n_param, fname, position);
+  if (lv == Sql_condition::WARN_LEVEL_ERROR)
+    my_error(code, MYF(0), n_param, fname, position);
+  else
+    push_warning_printf(thd, lv, code, ER_THD(thd, code),
+                        n_param, fname, position);
 }
 
 
-
 /*
   Checks if the path has '.*' '[*]' or '**' constructions
   and sets the NO_WILDCARD_ALLOWED error if the case.
@@ -506,11 +545,71 @@ longlong Item_func_json_valid::val_int()
 }
 
 
+bool Item_func_json_equals::fix_length_and_dec()
+{
+  if (Item_bool_func::fix_length_and_dec())
+    return TRUE;
+  set_maybe_null();
+  return FALSE;
+}
+
+
+longlong Item_func_json_equals::val_int()
+{
+  longlong result= 0;
+
+  String a_tmp, b_tmp;
+
+  String *a= args[0]->val_json(&a_tmp);
+  String *b= args[1]->val_json(&b_tmp);
+
+  DYNAMIC_STRING a_res;
+  if (init_dynamic_string(&a_res, NULL, 0, 0))
+  {
+    null_value= 1;
+    return 1;
+  }
+
+  DYNAMIC_STRING b_res;
+  if (init_dynamic_string(&b_res, NULL, 0, 0))
+  {
+    dynstr_free(&a_res);
+    null_value= 1;
+    return 1;
+  }
+
+  if ((null_value= args[0]->null_value || args[1]->null_value))
+  {
+    null_value= 1;
+    goto end;
+  }
+
+  if (json_normalize(&a_res, a->ptr(), a->length(), a->charset()))
+  {
+    null_value= 1;
+    goto end;
+  }
+
+  if (json_normalize(&b_res, b->ptr(), b->length(), b->charset()))
+  {
+    null_value= 1;
+    goto end;
+  }
+
+  result= strcmp(a_res.str, b_res.str) ? 0 : 1;
+
+end:
+  dynstr_free(&b_res);
+  dynstr_free(&a_res);
+  return result;
+}
+
+
 bool Item_func_json_exists::fix_length_and_dec()
 {
   if (Item_bool_func::fix_length_and_dec())
     return TRUE;
-  maybe_null= 1;
+  set_maybe_null();
   path.set_constant_flag(args[1]->const_item());
   return FALSE;
 }
@@ -564,7 +663,7 @@ bool Item_func_json_value::fix_length_and_dec()
   collation.set(args[0]->collation);
   max_length= args[0]->max_length;
   set_constant_flag(args[1]->const_item());
-  maybe_null= 1;
+  set_maybe_null();
   return FALSE;
 }
 
@@ -574,7 +673,7 @@ bool Item_func_json_query::fix_length_and_dec()
   collation.set(args[0]->collation);
   max_length= args[0]->max_length;
   set_constant_flag(args[1]->const_item());
-  maybe_null= 1;
+  set_maybe_null();
   return FALSE;
 }
 
@@ -703,9 +802,9 @@ String *Item_func_json_quote::val_str(String *str)
   str->length(0);
   str->set_charset(&my_charset_utf8mb4_bin);
 
-  if (str->append("\"", 1) ||
+  if (str->append('"') ||
       st_append_escaped(str, s) ||
-      str->append("\"", 1))
+      str->append('"'))
   {
     /* Report an error. */
     null_value= 1;
@@ -721,7 +820,7 @@ bool Item_func_json_unquote::fix_length_and_dec()
   collation.set(&my_charset_utf8mb3_general_ci,
                 DERIVATION_COERCIBLE, MY_REPERTOIRE_ASCII);
   max_length= args[0]->max_length;
-  maybe_null= 1;
+  set_maybe_null();
   return FALSE;
 }
 
@@ -842,7 +941,7 @@ bool Item_func_json_extract::fix_length_and_dec()
   max_length= args[0]->max_length * (arg_count - 1);
 
   mark_constant_paths(paths, args+1, arg_count-1);
-  maybe_null= 1;
+  set_maybe_null();
   return FALSE;
 }
 
@@ -917,7 +1016,7 @@ String *Item_func_json_extract::read_json(String *str,
     str->set_charset(js->charset());
     str->length(0);
 
-    if (possible_multiple_values && str->append("[", 1))
+    if (possible_multiple_values && str->append('['))
       goto error;
   }
 
@@ -965,7 +1064,7 @@ String *Item_func_json_extract::read_json(String *str,
     if (!possible_multiple_values)
     {
       /* Loop to the end of the JSON just to make sure it's valid. */
-      while (json_get_path_next(&je, &p) == 0) {}
+      while (json_scan_next(&je) == 0) {}
       break;
     }
   }
@@ -979,14 +1078,12 @@ String *Item_func_json_extract::read_json(String *str,
     goto return_null;
   }
 
-  if (possible_multiple_values && str->append("]", 1))
+  if (possible_multiple_values && str->append(']'))
     goto error; /* Out of memory. */
 
   js= str;
   json_scan_start(&je, js->charset(),(const uchar *) js->ptr(),
                   (const uchar *) js->ptr() + js->length());
-  tmp_js.length(0);
-  tmp_js.set_charset(js->charset());
   if (json_nice(&je, &tmp_js, Item_func_json_format::LOOSE))
     goto error;
 
@@ -1097,7 +1194,7 @@ my_decimal *Item_func_json_extract::val_decimal(my_decimal *to)
       case JSON_VALUE_OBJECT:
       case JSON_VALUE_ARRAY:
       case JSON_VALUE_FALSE:
-      case JSON_VALUE_UNINITALIZED:
+      case JSON_VALUE_UNINITIALIZED:
       case JSON_VALUE_NULL:
         int2my_decimal(E_DEC_FATAL_ERROR, 0, false/*unsigned_flag*/, to);
         return to;
@@ -1113,7 +1210,7 @@ bool Item_func_json_contains::fix_length_and_dec()
 {
   a2_constant= args[1]->const_item();
   a2_parsed= FALSE;
-  maybe_null= 1;
+  set_maybe_null();
   if (arg_count > 2)
     path.set_constant_flag(args[2]->const_item());
   return Item_bool_func::fix_length_and_dec();
@@ -1246,7 +1343,7 @@ static int check_contains(json_engine_t *js, json_engine_t *value)
     if (value->value_type != JSON_VALUE_STRING)
       return FALSE;
     /*
-       TODO: make proper json-json comparison here that takes excapint
+       TODO: make proper json-json comparison here that takes excipient
              into account.
      */
     return value->value_len == js->value_len &&
@@ -1372,7 +1469,7 @@ bool Item_func_json_contains_path::fix_length_and_dec()
 {
   ooa_constant= args[1]->const_item();
   ooa_parsed= FALSE;
-  maybe_null= 1;
+  set_maybe_null();
   mark_constant_paths(paths, args+2, arg_count-2);
   return Item_bool_func::fix_length_and_dec();
 }
@@ -1534,7 +1631,7 @@ longlong Item_func_json_contains_path::val_int()
     n_found= arg_count - 2;
   }
   else
-    n_found= 0; /* Jost to prevent 'uninitialized value' warnings */
+    n_found= 0; /* Just to prevent 'uninitialized value' warnings */
 
   result= 0;
   while (json_get_path_next(&je, &p) == 0)
@@ -1584,14 +1681,14 @@ null_return:
   `CONVERT(arg USING charset)` is actually a general purpose string
   expression, not a JSON expression.
 */
-static bool is_json_type(const Item *item)
+bool is_json_type(const Item *item)
 {
   for ( ; ; )
   {
     if (Type_handler_json_common::is_json_type_handler(item->type_handler()))
       return true;
     const Item_func_conv_charset *func;
-    if (!(func= dynamic_cast<const Item_func_conv_charset*>(item)))
+    if (!(func= dynamic_cast<const Item_func_conv_charset*>(item->real_item())))
       return false;
     item= func->arguments()[0];
   }
@@ -1632,15 +1729,15 @@ static int append_json_value(String *str, Item *item, String *tmp_val)
 
     if (item->result_type() == STRING_RESULT)
     {
-      return str->append("\"", 1) ||
+      return str->append('"') ||
              st_append_escaped(str, sv) ||
-             str->append("\"", 1);
+             str->append('"');
     }
     return st_append_escaped(str, sv);
   }
 
 append_null:
-  return str->append("null", 4);
+  return str->append(STRING_WITH_LEN("null"));
 }
 
 
@@ -1678,15 +1775,15 @@ static int append_json_value_from_field(String *str,
 
     if (i->result_type() == STRING_RESULT)
     {
-      return str->append("\"", 1) ||
+      return str->append('"') ||
              st_append_escaped(str, sv) ||
-             str->append("\"", 1);
+             str->append('"');
     }
     return st_append_escaped(str, sv);
   }
 
 append_null:
-  return str->append("null", 4);
+  return str->append(STRING_WITH_LEN("null"));
 }
 
 
@@ -1696,7 +1793,7 @@ static int append_json_keyname(String *str, Item *item, String *tmp_val)
   if (item->null_value)
     goto append_null;
 
-  return str->append("\"", 1) ||
+  return str->append('"') ||
          st_append_escaped(str, sv) ||
          str->append("\": ", 3);
 
@@ -1726,7 +1823,7 @@ bool Item_func_json_array::fix_length_and_dec()
     return TRUE;
 
   for (n_arg=0 ; n_arg < arg_count ; n_arg++)
-    char_length+= args[n_arg]->max_char_length() + 4;
+    char_length+= static_cast<ulonglong>(args[n_arg]->max_char_length()) + 4;
 
   fix_char_length_ulonglong(char_length);
   tmp_val.set_charset(collation.collation);
@@ -1736,13 +1833,13 @@ bool Item_func_json_array::fix_length_and_dec()
 
 String *Item_func_json_array::val_str(String *str)
 {
-  DBUG_ASSERT(fixed == 1);
+  DBUG_ASSERT(fixed());
   uint n_arg;
 
   str->length(0);
   str->set_charset(collation.collation);
 
-  if (str->append("[", 1) ||
+  if (str->append('[') ||
       ((arg_count > 0) && append_json_value(str, args[0], &tmp_val)))
     goto err_return;
 
@@ -1753,7 +1850,7 @@ String *Item_func_json_array::val_str(String *str)
       goto err_return;
   }
 
-  if (str->append("]", 1))
+  if (str->append(']'))
     goto err_return;
 
   if (result_limit == 0)
@@ -1785,11 +1882,12 @@ bool Item_func_json_array_append::fix_length_and_dec()
   for (n_arg= 1; n_arg < arg_count; n_arg+= 2)
   {
     paths[n_arg/2].set_constant_flag(args[n_arg]->const_item());
-    char_length+= args[n_arg/2+1]->max_char_length() + 4;
+    char_length+=
+        static_cast<ulonglong>(args[n_arg+1]->max_char_length()) + 4;
   }
 
   fix_char_length_ulonglong(char_length);
-  maybe_null= 1;
+  set_maybe_null();
   return FALSE;
 }
 
@@ -1801,8 +1899,9 @@ String *Item_func_json_array_append::val_str(String *str)
   uint n_arg, n_path;
   size_t str_rest_len;
   const uchar *ar_end;
+  THD *thd= current_thd;
 
-  DBUG_ASSERT(fixed == 1);
+  DBUG_ASSERT(fixed());
 
   if ((null_value= args[0]->null_value))
     return 0;
@@ -1828,6 +1927,7 @@ String *Item_func_json_array_append::val_str(String *str)
 
     json_scan_start(&je, js->charset(),(const uchar *) js->ptr(),
                     (const uchar *) js->ptr() + js->length());
+    je.killed_ptr= (uchar*)&thd->killed;
 
     c_path->cur_step= c_path->p.steps;
 
@@ -1882,11 +1982,11 @@ String *Item_func_json_array_append::val_str(String *str)
       else
         c_to= je.value_end;
 
-      if (str->append("[", 1) ||
+      if (str->append('[') ||
           str->append((const char *) c_from, c_to - c_from) ||
           str->append(", ", 2) ||
           append_json_value(str, args[n_arg+1], &tmp_val) ||
-          str->append("]", 1) ||
+          str->append(']') ||
           str->append((const char *) je.s.c_str,
                       js->end() - (const char *) je.s.c_str))
         goto return_null; /* Out of memory. */
@@ -1908,8 +2008,7 @@ String *Item_func_json_array_append::val_str(String *str)
 
   json_scan_start(&je, js->charset(),(const uchar *) js->ptr(),
                   (const uchar *) js->ptr() + js->length());
-  str->length(0);
-  str->set_charset(js->charset());
+  je.killed_ptr= (uchar*)&thd->killed;
   if (json_nice(&je, str, Item_func_json_format::LOOSE))
     goto js_error;
 
@@ -1919,6 +2018,7 @@ js_error:
   report_json_error(js, &je, 0);
 
 return_null:
+  thd->check_killed(); // to get the error message right
   null_value= 1;
   return 0;
 }
@@ -1929,8 +2029,9 @@ String *Item_func_json_array_insert::val_str(String *str)
   json_engine_t je;
   String *js= args[0]->val_json(&tmp_js);
   uint n_arg, n_path;
+  THD *thd= current_thd;
 
-  DBUG_ASSERT(fixed == 1);
+  DBUG_ASSERT(fixed());
 
   if ((null_value= args[0]->null_value))
     return 0;
@@ -1966,6 +2067,7 @@ String *Item_func_json_array_insert::val_str(String *str)
 
     json_scan_start(&je, js->charset(),(const uchar *) js->ptr(),
                     (const uchar *) js->ptr() + js->length());
+    je.killed_ptr= (uchar*)&thd->killed;
 
     c_path->cur_step= c_path->p.steps;
 
@@ -2005,7 +2107,7 @@ String *Item_func_json_array_insert::val_str(String *str)
         goto js_error;
     }
 
-    if (unlikely(je.s.error))
+    if (unlikely(je.s.error || *je.killed_ptr))
       goto js_error;
 
     str->length(0);
@@ -2049,8 +2151,7 @@ String *Item_func_json_array_insert::val_str(String *str)
 
   json_scan_start(&je, js->charset(),(const uchar *) js->ptr(),
                   (const uchar *) js->ptr() + js->length());
-  str->length(0);
-  str->set_charset(js->charset());
+  je.killed_ptr= (uchar*)&thd->killed;
   if (json_nice(&je, str, Item_func_json_format::LOOSE))
     goto js_error;
 
@@ -2059,6 +2160,7 @@ String *Item_func_json_array_insert::val_str(String *str)
 js_error:
   report_json_error(js, &je, 0);
 return_null:
+  thd->check_killed(); // to get the error message right
   null_value= 1;
   return 0;
 }
@@ -2066,13 +2168,13 @@ return_null:
 
 String *Item_func_json_object::val_str(String *str)
 {
-  DBUG_ASSERT(fixed == 1);
+  DBUG_ASSERT(fixed());
   uint n_arg;
 
   str->length(0);
   str->set_charset(collation.collation);
 
-  if (str->append("{", 1) ||
+  if (str->append('{') ||
       (arg_count > 0 &&
        (append_json_keyname(str, args[0], &tmp_val) ||
         append_json_value(str, args[1], &tmp_val))))
@@ -2086,7 +2188,7 @@ String *Item_func_json_object::val_str(String *str)
       goto err_return;
   }
 
-  if (str->append("}", 1))
+  if (str->append('}'))
     goto err_return;
 
   if (result_limit == 0)
@@ -2133,7 +2235,7 @@ static int do_merge(String *str, json_engine_t *je1, json_engine_t *je2)
   
     json_string_set_cs(&key_name, je1->s.cs);
 
-    if (str->append("{", 1))
+    if (str->append('{'))
       return 3;
     while (json_scan_next(je1) == 0 &&
            je1->state != JST_OBJ_END)
@@ -2159,7 +2261,7 @@ static int do_merge(String *str, json_engine_t *je1, json_engine_t *je2)
         *je2= sav_je2;
       }
 
-      if (str->append("\"", 1) ||
+      if (str->append('"') ||
           append_simple(str, key_start, key_end - key_start) ||
           str->append("\":", 2))
         return 3;
@@ -2227,8 +2329,7 @@ merged_j1:
             return 2;
           continue;
         }
-        if (json_skip_key(je2) ||
-            json_skip_level(je1))
+        if (json_skip_key(je2) || json_skip_level(je1))
           return 1;
         goto continue_j2;
       }
@@ -2244,7 +2345,7 @@ merged_j1:
       if (json_skip_key(je2))
         return 1;
 
-      if (str->append("\"", 1) ||
+      if (str->append('"') ||
           append_simple(str, key_start, je2->s.c_str - key_start))
         return 3;
 
@@ -2252,7 +2353,7 @@ continue_j2:
       continue;
     }
 
-    if (str->append("}", 1))
+    if (str->append('}'))
       return 3;
   }
   else
@@ -2272,7 +2373,7 @@ continue_j2:
     }
     else
     {
-      if (str->append("[", 1))
+      if (str->append('['))
         return 3;
       if (je1->value_type == JSON_VALUE_OBJECT)
       {
@@ -2314,7 +2415,7 @@ continue_j2:
       return 3;
 
     if (je2->value_type != JSON_VALUE_ARRAY &&
-        str->append("]", 1))
+        str->append(']'))
       return 3;
   }
 
@@ -2324,10 +2425,11 @@ continue_j2:
 
 String *Item_func_json_merge::val_str(String *str)
 {
-  DBUG_ASSERT(fixed == 1);
+  DBUG_ASSERT(fixed());
   json_engine_t je1, je2;
   String *js1= args[0]->val_json(&tmp_js1), *js2=NULL;
   uint n_arg;
+  THD *thd= current_thd;
   LINT_INIT(js2);
 
   if (args[0]->null_value)
@@ -2344,9 +2446,11 @@ String *Item_func_json_merge::val_str(String *str)
 
     json_scan_start(&je1, js1->charset(),(const uchar *) js1->ptr(),
                     (const uchar *) js1->ptr() + js1->length());
+    je1.killed_ptr= (uchar*)&thd->killed;
 
     json_scan_start(&je2, js2->charset(),(const uchar *) js2->ptr(),
                     (const uchar *) js2->ptr() + js2->length());
+    je2.killed_ptr= (uchar*)&thd->killed;
 
     if (do_merge(str, &je1, &je2))
       goto error_return;
@@ -2368,8 +2472,7 @@ String *Item_func_json_merge::val_str(String *str)
 
   json_scan_start(&je1, js1->charset(),(const uchar *) js1->ptr(),
                   (const uchar *) js1->ptr() + js1->length());
-  str->length(0);
-  str->set_charset(js1->charset());
+  je1.killed_ptr= (uchar*)&thd->killed;
   if (json_nice(&je1, str, Item_func_json_format::LOOSE))
     goto error_return;
 
@@ -2381,6 +2484,7 @@ error_return:
     report_json_error(js1, &je1, 0);
   if (je2.s.error)
     report_json_error(js2, &je2, n_arg);
+  thd->check_killed(); // to get the error message right
 null_return:
   null_value= 1;
   return NULL;
@@ -2413,7 +2517,7 @@ static int copy_value_patch(String *str, json_engine_t *je)
   }
   /* JSON_VALUE_OBJECT */
 
-  if (str->append("{", 1))
+  if (str->append('{'))
     return 1;
   while (json_scan_next(je) == 0 && je->state != JST_OBJ_END)
   {
@@ -2436,12 +2540,12 @@ static int copy_value_patch(String *str, json_engine_t *je)
     else
       first_key= 0;
 
-    if (str->append("\"", 1) ||
+    if (str->append('"') ||
         append_simple(str, key_start, je->value_begin - key_start) ||
         copy_value_patch(str, je))
       return 1;
   }
-  if (str->append("}", 1))
+  if (str->append('}'))
     return 1;
 
   return 0;
@@ -2477,7 +2581,7 @@ static int do_merge_patch(String *str, json_engine_t *je1, json_engine_t *je2,
     *empty_result= FALSE;
     json_string_set_cs(&key_name, je1->s.cs);
 
-    if (str->append("{", 1))
+    if (str->append('{'))
       return 3;
     while (json_scan_next(je1) == 0 &&
            je1->state != JST_OBJ_END)
@@ -2503,7 +2607,7 @@ static int do_merge_patch(String *str, json_engine_t *je1, json_engine_t *je2,
         *je2= sav_je2;
       }
 
-      if (str->append("\"", 1) ||
+      if (str->append('"') ||
           append_simple(str, key_start, key_end - key_start) ||
           str->append("\":", 2))
         return 3;
@@ -2594,7 +2698,7 @@ merged_j1:
       if (!first_key && str->append(", ", 2))
         return 3;
 
-      if (str->append("\"", 1) ||
+      if (str->append('"') ||
           append_simple(str, key_start, key_end - key_start) ||
           str->append("\":", 2))
         return 3;
@@ -2615,7 +2719,7 @@ continue_j2:
       continue;
     }
 
-    if (str->append("}", 1))
+    if (str->append('}'))
       return 3;
   }
   else
@@ -2634,11 +2738,12 @@ continue_j2:
 
 String *Item_func_json_merge_patch::val_str(String *str)
 {
-  DBUG_ASSERT(fixed == 1);
+  DBUG_ASSERT(fixed());
   json_engine_t je1, je2;
   String *js1= args[0]->val_json(&tmp_js1), *js2=NULL;
   uint n_arg;
   bool empty_result, merge_to_null;
+  THD *thd= current_thd;
 
   /* To report errors properly if some JSON is invalid. */
   je1.s.error= je2.s.error= 0;
@@ -2655,6 +2760,7 @@ String *Item_func_json_merge_patch::val_str(String *str)
 
     json_scan_start(&je2, js2->charset(),(const uchar *) js2->ptr(),
                     (const uchar *) js2->ptr() + js2->length());
+    je2.killed_ptr= (uchar*)&thd->killed;
 
     if (merge_to_null)
     {
@@ -2674,12 +2780,13 @@ String *Item_func_json_merge_patch::val_str(String *str)
 
     json_scan_start(&je1, js1->charset(),(const uchar *) js1->ptr(),
                     (const uchar *) js1->ptr() + js1->length());
+    je1.killed_ptr= (uchar*)&thd->killed;
 
     if (do_merge_patch(str, &je1, &je2, &empty_result))
       goto error_return;
 
     if (empty_result)
-      str->append("null");
+      str->append(STRING_WITH_LEN("null"));
 
 cont_point:
     {
@@ -2702,8 +2809,7 @@ cont_point:
 
   json_scan_start(&je1, js1->charset(),(const uchar *) js1->ptr(),
                   (const uchar *) js1->ptr() + js1->length());
-  str->length(0);
-  str->set_charset(js1->charset());
+  je1.killed_ptr= (uchar*)&thd->killed;
   if (json_nice(&je1, str, Item_func_json_format::LOOSE))
     goto error_return;
 
@@ -2715,6 +2821,7 @@ error_return:
     report_json_error(js1, &je1, 0);
   if (je2.s.error)
     report_json_error(js2, &je2, n_arg);
+  thd->check_killed(); // to get the error message right
 null_return:
   null_value= 1;
   return NULL;
@@ -2725,7 +2832,7 @@ bool Item_func_json_length::fix_length_and_dec()
 {
   if (arg_count > 1)
     path.set_constant_flag(args[1]->const_item());
-  maybe_null= 1;
+  set_maybe_null();
   max_length= 10;
   return FALSE;
 }
@@ -2870,8 +2977,8 @@ longlong Item_func_json_depth::val_int()
 bool Item_func_json_type::fix_length_and_dec()
 {
   collation.set(&my_charset_utf8mb3_general_ci);
-  max_length= 12;
-  maybe_null= 1;
+  max_length= 12 * collation.collation->mbmaxlen;
+  set_maybe_null();
   return FALSE;
 }
 
@@ -2936,11 +3043,12 @@ bool Item_func_json_insert::fix_length_and_dec()
   for (n_arg= 1; n_arg < arg_count; n_arg+= 2)
   {
     paths[n_arg/2].set_constant_flag(args[n_arg]->const_item());
-    char_length+= args[n_arg/2+1]->max_char_length() + 4;
+    char_length+=
+        static_cast<ulonglong>(args[n_arg+1]->max_char_length()) + 4;
   }
 
   fix_char_length_ulonglong(char_length);
-  maybe_null= 1;
+  set_maybe_null();
   return FALSE;
 }
 
@@ -2951,8 +3059,9 @@ String *Item_func_json_insert::val_str(String *str)
   String *js= args[0]->val_json(&tmp_js);
   uint n_arg, n_path;
   json_string_t key_name;
+  THD *thd= current_thd;
 
-  DBUG_ASSERT(fixed == 1);
+  DBUG_ASSERT(fixed());
 
   if ((null_value= args[0]->null_value))
     return 0;
@@ -2991,6 +3100,7 @@ String *Item_func_json_insert::val_str(String *str)
 
     json_scan_start(&je, js->charset(),(const uchar *) js->ptr(),
                     (const uchar *) js->ptr() + js->length());
+    je.killed_ptr= (uchar*)&thd->killed;
 
     if (c_path->p.last_step < c_path->p.steps)
       goto v_found;
@@ -3040,7 +3150,7 @@ String *Item_func_json_insert::val_str(String *str)
         str->length(0);
         /* Wrap the value as an array. */
         if (append_simple(str, js->ptr(), (const char *) v_from - js->ptr()) ||
-            (do_array_autowrap && str->append("[", 1)))
+            (do_array_autowrap && str->append('[')))
           goto js_error; /* Out of memory. */
 
         if (je.value_type == JSON_VALUE_OBJECT)
@@ -3053,7 +3163,7 @@ String *Item_func_json_insert::val_str(String *str)
              (append_simple(str, v_from, je.s.c_str - v_from) ||
               str->append(", ", 2))) ||
             append_json_value(str, args[n_arg+1], &tmp_val) ||
-            (do_array_autowrap && str->append("]", 1)) ||
+            (do_array_autowrap && str->append(']')) ||
             append_simple(str, je.s.c_str, js->end()-(const char *) je.s.c_str))
           goto js_error; /* Out of memory. */
 
@@ -3124,7 +3234,7 @@ String *Item_func_json_insert::val_str(String *str)
       str->length(0);
       if (append_simple(str, js->ptr(), v_to - js->ptr()) ||
           (n_key > 0 && str->append(", ", 2)) ||
-          str->append("\"", 1) ||
+          str->append('"') ||
           append_simple(str, lp->key, lp->key_end - lp->key) ||
           str->append("\":", 2) ||
           append_json_value(str, args[n_arg+1], &tmp_val) ||
@@ -3172,7 +3282,7 @@ continue_point:
 
   json_scan_start(&je, js->charset(),(const uchar *) js->ptr(),
                   (const uchar *) js->ptr() + js->length());
-  str->length(0);
+  je.killed_ptr= (uchar*)&thd->killed;
   if (json_nice(&je, str, Item_func_json_format::LOOSE))
     goto js_error;
 
@@ -3180,6 +3290,7 @@ continue_point:
 
 js_error:
   report_json_error(js, &je, 0);
+  thd->check_killed(); // to get the error message right
 return_null:
   null_value= 1;
   return 0;
@@ -3192,7 +3303,7 @@ bool Item_func_json_remove::fix_length_and_dec()
   max_length= args[0]->max_length;
 
   mark_constant_paths(paths, args+1, arg_count-1);
-  maybe_null= 1;
+  set_maybe_null();
   return FALSE;
 }
 
@@ -3203,8 +3314,9 @@ String *Item_func_json_remove::val_str(String *str)
   String *js= args[0]->val_json(&tmp_js);
   uint n_arg, n_path;
   json_string_t key_name;
+  THD *thd= current_thd;
 
-  DBUG_ASSERT(fixed == 1);
+  DBUG_ASSERT(fixed());
 
   if (args[0]->null_value)
     goto null_return;
@@ -3249,6 +3361,7 @@ String *Item_func_json_remove::val_str(String *str)
 
     json_scan_start(&je, js->charset(),(const uchar *) js->ptr(),
                     (const uchar *) js->ptr() + js->length());
+    je.killed_ptr= (uchar*)&thd->killed;
 
     c_path->cur_step= c_path->p.steps;
 
@@ -3357,8 +3470,7 @@ v_found:
 
   json_scan_start(&je, js->charset(),(const uchar *) js->ptr(),
                   (const uchar *) js->ptr() + js->length());
-  str->length(0);
-  str->set_charset(js->charset());
+  je.killed_ptr= (uchar*)&thd->killed;
   if (json_nice(&je, str, Item_func_json_format::LOOSE))
     goto js_error;
 
@@ -3366,6 +3478,7 @@ v_found:
   return str;
 
 js_error:
+  thd->check_killed(); // to get the error message right
   report_json_error(js, &je, 0);
 null_return:
   null_value= 1;
@@ -3377,7 +3490,7 @@ bool Item_func_json_keys::fix_length_and_dec()
 {
   collation.set(args[0]->collation);
   max_length= args[0]->max_length;
-  maybe_null= 1;
+  set_maybe_null();
   if (arg_count > 1)
     path.set_constant_flag(args[1]->const_item());
   return FALSE;
@@ -3469,7 +3582,7 @@ skip_search:
     goto null_return;
   
   str->length(0);
-  if (str->append("[", 1))
+  if (str->append('['))
     goto err_return; /* Out of memory. */
   /* Parse the OBJECT collecting the keys. */
   while (json_scan_next(&je) == 0 && je.state != JST_OBJ_END)
@@ -3492,9 +3605,9 @@ skip_search:
       if (!check_key_in_list(str, key_start, key_len))
       { 
         if ((n_keys > 0 && str->append(", ", 2)) ||
-          str->append("\"", 1) ||
+          str->append('"') ||
           append_simple(str, key_start, key_len) ||
-          str->append("\"", 1))
+          str->append('"'))
         goto err_return;
         n_keys++;
       }
@@ -3509,7 +3622,7 @@ skip_search:
     }
   }
 
-  if (unlikely(je.s.error || str->append("]", 1)))
+  if (unlikely(je.s.error || str->append(']')))
     goto err_return;
 
   null_value= 0;
@@ -3547,7 +3660,7 @@ bool Item_func_json_search::fix_length_and_dec()
 
   /*
     It's rather difficult to estimate the length of the result.
-    I belive arglen^2 is the reasonable upper limit.
+    I believe arglen^2 is the reasonable upper limit.
   */
   if (args[0]->max_length > SQR_MAX_BLOB_WIDTH)
     max_length= MAX_BLOB_WIDTH;
@@ -3562,7 +3675,7 @@ bool Item_func_json_search::fix_length_and_dec()
 
   if (arg_count > 4)
     mark_constant_paths(paths, args+4, arg_count-4);
-  maybe_null= 1;
+  set_maybe_null();
   return FALSE;
 }
 
@@ -3613,14 +3726,14 @@ static int append_json_path(String *str, const json_path_t *p)
     else /*JSON_PATH_ARRAY*/
     {
 
-      if (str->append("[", 1) ||
+      if (str->append('[') ||
           str->append_ulonglong(c->n_item) ||
-          str->append("]", 1))
+          str->append(']'))
         return TRUE;
     }
   }
 
-  return str->append("\"", 1);
+  return str->append('"');
 }
 
 
@@ -3681,7 +3794,7 @@ String *Item_func_json_search::val_str(String *str)
         {
           if (n_path_found == 2)
           {
-            if (str->append("[", 1) ||
+            if (str->append('[') ||
                 append_json_path(str, &sav_path))
                 goto js_error;
           }
@@ -3707,7 +3820,7 @@ end:
   }
   else
   {
-    if (str->append("]", 1))
+    if (str->append(']'))
       goto js_error;
   }
 
@@ -3723,21 +3836,21 @@ null_return:
 }
 
 
-const char *Item_func_json_format::func_name() const
+LEX_CSTRING Item_func_json_format::func_name_cstring() const
 {
   switch (fmt)
   {
   case COMPACT:
-    return "json_compact";
+    return { STRING_WITH_LEN("json_compact") };
   case LOOSE:
-    return "json_loose";
+    return { STRING_WITH_LEN("json_loose") };
   case DETAILED:
-    return "json_detailed";
+    return { STRING_WITH_LEN("json_detailed") };
   default:
     DBUG_ASSERT(0);
   };
 
-  return "";
+  return NULL_clex_str;
 }
 
 
@@ -3746,7 +3859,7 @@ bool Item_func_json_format::fix_length_and_dec()
   decimals= 0;
   collation.set(args[0]->collation);
   max_length= args[0]->max_length;
-  maybe_null= 1;
+  set_maybe_null();
   return FALSE;
 }
 
@@ -3756,6 +3869,7 @@ String *Item_func_json_format::val_str(String *str)
   String *js= args[0]->val_json(&tmp_js);
   json_engine_t je;
   int tab_size= 4;
+  THD *thd= current_thd;
 
   if ((null_value= args[0]->null_value))
     return 0;
@@ -3779,13 +3893,13 @@ String *Item_func_json_format::val_str(String *str)
 
   json_scan_start(&je, js->charset(), (const uchar *) js->ptr(),
                   (const uchar *) js->ptr()+js->length());
+  je.killed_ptr= (uchar*)&thd->killed;
 
-  str->length(0);
-  str->set_charset(js->charset());
   if (json_nice(&je, str, fmt, tab_size))
   {
     null_value= 1;
     report_json_error(js, &je, 0);
+    thd->check_killed(); // to get the error message right
     return 0;
   }
 
@@ -3945,7 +4059,7 @@ Item_func_json_objectagg(THD *thd, Item_func_json_objectagg *item)
 {
   quick_group= FALSE;
   result.set_charset(collation.collation);
-  result.append("{");
+  result.append('{');
 }
 
 
@@ -3953,14 +4067,14 @@ bool
 Item_func_json_objectagg::fix_fields(THD *thd, Item **ref)
 {
   uint i;                       /* for loop variable */
-  DBUG_ASSERT(fixed == 0);
+  DBUG_ASSERT(fixed() == 0);
 
   memcpy(orig_args, args, sizeof(Item*) * arg_count);
 
   if (init_sum_func_check(thd))
     return TRUE;
 
-  maybe_null= 1;
+  set_maybe_null();
 
   /*
     Fix fields for select list and ORDER clause
@@ -3970,9 +4084,7 @@ Item_func_json_objectagg::fix_fields(THD *thd, Item **ref)
   {
     if (args[i]->fix_fields_if_needed_for_scalar(thd, &args[i]))
       return TRUE;
-    m_with_subquery|= args[i]->with_subquery();
-    with_param|= args[i]->with_param;
-    with_window_func|= args[i]->with_window_func;
+    with_flags|= args[i]->with_flags;
   }
 
   /* skip charset aggregation for order columns */
@@ -3989,7 +4101,7 @@ Item_func_json_objectagg::fix_fields(THD *thd, Item **ref)
   if (check_sum_func(thd, ref))
     return TRUE;
 
-  fixed= 1;
+  base_flags|= item_base_t::FIXED;
   return FALSE;
 }
 
@@ -4028,11 +4140,11 @@ bool Item_func_json_objectagg::add()
 
   null_value= 0;
   if (result.length() > 1)
-    result.append(", ");
+    result.append(STRING_WITH_LEN(", "));
 
-  result.append("\"");
-  result.append(*key);
-  result.append("\":");
+  result.append('"');
+  st_append_escaped(&result,key);
+  result.append(STRING_WITH_LEN("\":"));
 
   buf.length(0);
   append_json_value(&result, args[1], &buf);
@@ -4043,12 +4155,57 @@ bool Item_func_json_objectagg::add()
 
 String* Item_func_json_objectagg::val_str(String* str)
 {
-  DBUG_ASSERT(fixed == 1);
+  DBUG_ASSERT(fixed());
   if (null_value)
     return 0;
 
-  result.append("}");
+  result.append('}');
   return &result;
 }
 
 
+String *Item_func_json_normalize::val_str(String *buf)
+{
+  String tmp;
+  String *raw_json= args[0]->val_str(&tmp);
+
+  DYNAMIC_STRING normalized_json;
+  if (init_dynamic_string(&normalized_json, NULL, 0, 0))
+  {
+    null_value= 1;
+    return NULL;
+  }
+
+  null_value= args[0]->null_value;
+  if (null_value)
+    goto end;
+
+  if (json_normalize(&normalized_json,
+                     raw_json->ptr(), raw_json->length(),
+                     raw_json->charset()))
+  {
+    null_value= 1;
+    goto end;
+  }
+
+  buf->length(0);
+  if (buf->append(normalized_json.str, normalized_json.length))
+  {
+    null_value= 1;
+    goto end;
+  }
+
+end:
+  dynstr_free(&normalized_json);
+  return null_value ? NULL : buf;
+}
+
+
+bool Item_func_json_normalize::fix_length_and_dec()
+{
+  collation.set(&my_charset_utf8mb4_bin);
+  /* 0 becomes 0.0E0, thus one character becomes 5 chars */
+  fix_char_length_ulonglong((ulonglong) args[0]->max_char_length() * 5);
+  set_maybe_null();
+  return FALSE;
+}
